/* Copyright 2015 Jetperch LLC.  All rights reserved. */

/**
 * @file
 *
 * @brief First-in, first-out queue (not thread safe).
 */

/**
 * @ingroup embc
 * @defgroup fifo FIFO
 *
 * @brief First-in, First-out queue (not thread safe).
 *
 * @{
 */

#ifndef FIFO_H_
#define FIFO_H_

#include <stdint.h>
#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/**
 * @brief The type used to represent the FIFO size.
 *
 * Note that the FIFO implementation currently requires a signed value
 * to operate properly.
 */
typedef int fifo_size_t;

/**
 * @brief The FIFO instance.
 */
struct fifo8_s {
    fifo_size_t head;   ///< The head for writes.
    fifo_size_t tail;   ///< The tail for reads.
    uint8_t * buffer;   ///< The data buffer.
    size_t buffer_size; ///< The data buffer total size.
};


/**
 * @brief Allocate a new FIFO instance.
 *
 * @param fifo [out] The pointer to the new FIFO instance.
 * @param depth [in] The number of bytes in the the FIFO instance.  Due to
 *      the FIFO implementation, the actual full level is 1 byte less.
 * @return 0 or error code.
 */
int fifo8_alloc(struct fifo8_s ** fifo, fifo_size_t depth);

/**
 * @brief Free a FIFO instance created by fifo8_alloc().
 *
 * @param fifo [in] The FIFO instance.
 */
void fifo8_free(struct fifo8_s * fifo);

/**
 * @brief Write data into a FIFO instance.
 *
 * @param fifo [in] The FIFO instance.
 * @param data [in] The data to write.
 * @param size [in] The number of bytes in data.
 * @param actual_size [out] The actual number of bytes written.  The FIFO
 *      will write as many bytes as possible until the FIFO becomes full.
 * @return 0 on success, EMBC_ERROR_FULL on FIFO full (check actual_size)
 *      or an error code.
 */
int fifo8_write(struct fifo8_s * fifo, uint8_t const * data, fifo_size_t size, fifo_size_t * actual_size);

/**
 * @brief Write a single byte into a FIFO instance.
 *
 * @param fifo [in] The FIFO instance.
 * @param data [in] The data to write.
 * @return 0 on sucess, EMBC_ERROR_FULL on FIFO full (check actual_size)
 *      or an error code.
 */
int fifo8_write_u8(struct fifo8_s * fifo, uint8_t data);

/**
 * @brief Get the number of bytes in the FIFO.
 *
 * @param fifo [in] The FIFO instance.
 * @return The number of data bytes currently in the FIFO.  Note that
 *      this is NOT the total storage capacity of the FIFO.
 */
fifo_size_t fifo8_size(struct fifo8_s * fifo);

/**
 * @brief Check if the FIFO is empty.
 *
 * @param fifo [in] The FIFO instance.
 * @return 1 if the FIFO is empty, else 0.
 */
int fifo8_is_empty(struct fifo8_s * fifo);

/**
 * @brief Check if the FIFO is full.
 *
 * @param fifo [in] The FIFO instance.
 * @return 1 if the FIFO is full, else 0.
 */
int fifo8_is_full(struct fifo8_s * fifo);

/**
 * @brief Read data from the FIFO.
 *
 * @param fifo [in] The FIFO instance.
 * @param data [out] The data read from the FIFO.
 * @param size [in] The number of bytes to read from the FIFO.  Note that
 *      data must also be at least this size.
 * @param actual_size [out] The total number of bytes actually read from the
 *      FIFO which can be less than size when the FIFO becomes empty.
 * @return 0 on success, EMBC_ERROR_EMPTY on FIFO empty (check actual_size)
 *      or an error code.
 */
int fifo8_read(struct fifo8_s * fifo, uint8_t * data, fifo_size_t size, fifo_size_t * actual_size);

/**
 * @brief Read a byte from the FIFO.
 *
 * @param fifo [in] The FIFO instance.
 * @param data [out] The data read from the FIFO.
 * @return 0 on success, EMBC_ERROR_EMPTY on FIFO empty (check actual_size)
 *      or an error code.
 */
int fifo8_read_u8(struct fifo8_s * fifo, uint8_t * data);

/** @} */

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* FIFO_H_ */
