#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#if defined(WIN32) || defined (__MINGW32__)
#include <windows.h>
#endif

#include "ll_ifc_transport_pc.h"
#include "ll_ifc_ensemble.h"

#if defined(WIN32) || defined (__MINGW32__)
#define SLEEP(x)        Sleep(x * 1000)
#define MS_SLEEP(x)     Sleep(x)
#define DEFAULT_BAUDRATE		(CBR_115200)
#else
#define SLEEP(x)        sleep(x)
#define MS_SLEEP(x)     usleep(x * 1000)
#define DEFAULT_BAUDRATE		(115200)
#endif

// Build Date
#define BUILD_DATE (__BUILD_DATE);

// Build Number
#define VERSION_MAJOR                (1) // 0-255
#define VERSION_MINOR                (3) // 0-255
#define VERSION_TAG     (__BUILD_NUMBER) // 0-65535

#define SENTINEL				(255)

#define TIME_BUFFER_SIZE    80      // force time buffer string's size

// used later to lookup names for properties
typedef struct {
    ensemble_property_t id;
    char*   name;
} property_id_to_name_map;

static void print_ll_ifc_error(char *label, int32_t ret_val)
{
    if (ret_val < 0)
    {
        // Map Error code to NACK code
        if (ret_val >= -99 && ret_val <= -1)
        {
            ret_val = 0 - ret_val;
        }
        fprintf(stderr, "ERROR(%s): Host interface - ", label);
        switch(ret_val)
        {
            case LL_IFC_NACK_CMD_NOT_SUPPORTED:
                fprintf(stderr, "NACK received - Command not supported");
                break;
            case LL_IFC_NACK_INCORRECT_CHKSUM:
                fprintf(stderr, "NACK received - Incorrect Checksum");
                break;
            case LL_IFC_NACK_PAYLOAD_LEN_OOR:
                fprintf(stderr, "NACK received - Payload length out of range");
                break;
            case LL_IFC_NACK_PAYLOAD_OOR:
                fprintf(stderr, "NACK received - Payload out of range");
                break;
            case LL_IFC_NACK_BOOTUP_IN_PROGRESS:
                fprintf(stderr, "NACK received - Not allowed, bootup in progress");
                break;
            case LL_IFC_NACK_BUSY_TRY_AGAIN:
                fprintf(stderr, "NACK received - Busy try again");
                break;
            case LL_IFC_NACK_APP_TOKEN_REG:
                fprintf(stderr, "NACK received - Application Token not registered");
                break;
            case LL_IFC_NACK_PAYLOAD_LEN_EXCEEDED:
                fprintf(stderr, "NACK received - Payload length greater than maximum");
                break;
            case LL_IFC_NACK_NOT_IN_MAILBOX_MODE:
                fprintf(stderr, "NACK received - Module is not in DOWNLINK_MAILBOX mode");
                break;
            case LL_IFC_NACK_PAYLOAD_BAD_PROPERTY:
                fprintf(stderr, "NACK received - Invalid property specified");
                break;
            case LL_IFC_NACK_NODATA:
                fprintf(stderr, "NACK received - No data available");
                break;
            case LL_IFC_NACK_QUEUE_FULL:
                fprintf(stderr, "NACK received - Cannot accept data, queue is full");
                break;
            case LL_IFC_NACK_OTHER:
                fprintf(stderr, "NACK received - Other");
                break;
            case LL_IFC_ERROR_INCORRECT_PARAMETER:
                fprintf(stderr, "Invalid Parameter");
                break;
            case -103:
                fprintf(stderr, "Message Number mismatch");
                break;
            case -104:
                fprintf(stderr, "Checksum mismatch");
                break;
            case -105:
                fprintf(stderr, "Command mismatch");
                break;
            case -106:
                fprintf(stderr, "Timed out");
                break;
            case -107:
                fprintf(stderr, "Payload larger than buffer provided");
                break;
            default:
                fprintf(stderr, "Unknown Error %d", ret_val);
                break;

        }
        fprintf(stderr, "\n");
    }
}

static void print_ll_version(void)
{
    ll_version_t ver;
    int8_t i8_ret;

    i8_ret = ll_version_get(&ver);
    print_ll_ifc_error("ll_version_get", i8_ret);
    if (i8_ret == VERSION_LEN)
    {
        printf("Link Labs Version = %d.%d.%d\n", ver.major, ver.minor, ver.tag);
    }
}

static void print_uuid(void)
{
	uint64_t uuid = 0;
    int8_t ret;

	ret = ll_unique_id_get(&uuid);
    uuid &= 0xFFFFFFFFF;
    print_ll_ifc_error("ll_unique_id_get", ret);
	if (ret >= 0)
	{
        printf("Link Labs UUID = $301$0-0-0-%01x", (uint32_t)(uuid >> 32));
        printf("%08x\n", (uint32_t)uuid);
	}
}

static void print_ll_firmware_type(void)
{
    ll_firmware_type_t t;
    int8_t ret;

    ret = ll_firmware_type_get(&t);
    print_ll_ifc_error("ll_firmware_type", ret);

    if (ret == FIRMWARE_TYPE_LEN)
    {
        printf("Link Labs Firmware Type = %04d.%04d\n", t.cpu_code, t.functionality_code);
    }
}

static void print_ll_hardware_type(void)
{
    ll_hardware_type_t t;
    int8_t ret;

    ret = ll_hardware_type_get(&t);
    print_ll_ifc_error("ll_hardware_type", ret);

    if (ret == 1)
    {
        switch(t)
        {
            case UNAVAILABLE:
                break;
            case LLRLP20_V2:
                printf("Link Labs Hardware Type = LLRLP20 v2\n");
                break;
            case LLRXR26_V2:
                printf("Link Labs Hardware Type = LLRXR26 v2\n");
                break;
            case LLRLP20_V3:
                printf("Link Labs Hardware Type = LLRLP20 v3\n");
                break;
            case LLRXR26_V3:
                printf("Link Labs Hardware Type = LLRXR26 v3\n");
                break;
            default:
                break;
        }
    }
}

int32_t buffer_to_hex(char* buf)
{
    // only valid characters are 0:9,a:f,a:f
    uint16_t i;
    uint16_t num_chars = strlen(buf);

                    // odd length string check
    if (num_chars & 0x01)
    {
        return -1;
    }

    for(i=0; i<num_chars; i++)
    {
        char c = toupper(buf[i]);   // convert to all uppercase
        uint8_t x;

        // 0:9
        if (isdigit(c))
        {
            x = c-'0';
        }
        else if ((c >= 'A') && (c <= 'F'))
        {
            x = c-'A'+10;
        }
        else            // invalid character check
        {
            return -2;
        }

        // pack the buffer
        if (i & 0x01)
        {
            buf[(i-1)/2] |= x;
        }
        else
        {
            buf[i/2] = x<<4;
        }
    }
                        // null terminate the string
    buf[num_chars/2] = '\0';
    return num_chars/2;
}

static  property_id_to_name_map property_map[ENSEMBLE_PROP_MAX] = {
    { ENSEMBLE_PROP_FREQUENCY,             "Frequency" },
    { ENSEMBLE_PROP_BW,                    "Bandwidth" },
    { ENSEMBLE_PROP_SF,                    "Spreading Factor" },
    { ENSEMBLE_PROP_CR,                    "Coding Rate" },
    { ENSEMBLE_PROP_NODE_ROLE,             "Node Role" },
    { ENSEMBLE_PROP_TX_POWER,              "Power Level" },
    { ENSEMBLE_PROP_PREAMBLE_LENGTH,       "Preamble Length (ms)" },
    { ENSEMBLE_PROP_MAX_TX_ATTEMPTS,       "Max Tx Attempts Per Message" },
    { ENSEMBLE_PROP_MAX_PAYLOAD_SIZE,      "Max Payload Size" },
    { ENSEMBLE_PROP_GATEWAY_MAX_MSG_CNT,   "Max Gateway Buffered Msg Count" },
    { ENSEMBLE_PROP_TX_FAIL_RETRY_DELAY,   "Tx Retry Delay" },
    { ENSEMBLE_PROP_MSG_AVAIL_NOTIFY_MODE, "Host Notify Method" },
    { ENSEMBLE_PROP_MAIL_ACK_DELAY,        "Mail Acknowledge Delay" },
    { ENSEMBLE_PROP_REPEATER_AVAILABLE,    "Repeater Available" },
    { ENSEMBLE_PROP_MSG_TRAFFIC_KEY,       "Traffic Key" },                // its in the table but not to be accessed via the property interface
    };


static  char*   propertyNameToStr( ensemble_property_t property_id )
{
    int i;
                // we won't assume direct index-based LU is possible
    for( i = 0; i < ENSEMBLE_PROP_MAX; i++ )
    {
        if( property_map[i].id == property_id )
        {
            return property_map[i].name;
        }
    }
    return NULL;
}

static void print_irq_flags_text(uint32_t flags)
{
    // Most significant bit to least significant bit
    if(IRQ_FLAGS_ASSERT & flags)
    {
        printf("[IRQ_FLAGS_ASSERT]");
    }
    if(IRQ_FLAGS_APP_TOKEN_ERROR & flags)
    {
        printf("[IRQ_FLAGS_APP_TOKEN_ERROR]");
    }
    if(IRQ_FLAGS_CRYPTO_ERROR & flags)
    {
        printf("[IRQ_FLAGS_CRYPTO_ERROR]");
    }
    if(IRQ_FLAGS_DOWNLINK_REQUEST_ACK & flags)
    {
        printf("[IRQ_FLAGS_DOWNLINK_REQUEST_ACK]");
    }
    if(IRQ_FLAGS_APP_TOKEN_CONFIRMED & flags)
    {
        printf("[IRQ_FLAGS_APP_TOKEN_CONFIRMED]");
    }
    if(IRQ_FLAGS_CRYPTO_ESTABLISHED & flags)
    {
        printf("[IRQ_FLAGS_CRYPTO_ESTABLISHED]");
    }
    if(IRQ_FLAGS_DISCONNECTED & flags)
    {
        printf("[IRQ_FLAGS_DISCONNECTED]");
    }
    if(IRQ_FLAGS_CONNECTED & flags)
    {
        printf("[IRQ_FLAGS_CONNECTED]");
    }
    if(IRQ_FLAGS_RX_DONE & flags)
    {
        printf("[IRQ_FLAGS_RX_DONE]");
    }
    if(IRQ_FLAGS_TX_ERROR & flags)
    {
        printf("[IRQ_FLAGS_TX_ERROR]");
    }
    if(IRQ_FLAGS_TX_DONE & flags)
    {
        printf("[IRQ_FLAGS_TX_DONE]");
    }
    if(IRQ_FLAGS_ACK_RECEIVED & flags)
    {
        printf("[IRQ_FLAGS_ACK_RECEIVED]");
    }
    if(IRQ_FLAGS_RESET & flags)
    {
        printf("[IRQ_FLAGS_RESET]");
    }
    if(IRQ_FLAGS_WDOG_RESET & flags)
    {
        printf("[IRQ_FLAGS_WDOG_RESET]");
    }
    if(flags != 0)
    {
        printf("\n");
    }
}

static void usage(void)
{
    int i;

    printf("Usage:\n");
    printf("\n");
    printf("  --baudrate [-b] configure baudrate of the tty device, default: %d\n", LL_TTY_DEFAULT_BAUDRATE);
    printf("  --device [-D] choose tty device, default: "LL_TTY_DEFAULT_DEVICE"\n");
    printf("  --version [-v] present/return versioning information for the tool\n");
    printf("  --set [s] set the Ensemble MAC propery value specified as --property [-p]\n");
    printf("  --get [g] get the Ensemble MAC propery value specified as --property [-p]\n");
    printf("  --property [-p] property specified for get/set switches\n");
    printf("  --save [-w] store (write) radio settings to flash\n");
    printf("  --all [-a] dump all Ensemble MAC parameters for this device\n");
    printf("  --mac_set set MAC protocol\n");
    printf("  --mac_get get MAC protocol\n");
    printf("  --help [-h] print this help message\n");
    printf("  --uuid [-u] get unique identifier\n");
    printf("  --restore_defaults [-R] Restore default radio settings\n");
    printf("  --delete_settings [-d] delete saved settings from flash\n");
    printf("  --antenna_set Set the antenna configuration \n");
    printf("  --antenna_get Get the antenna configuration \n");
    printf("  --bootloader [-U] Reset to Bootloader mode (takes a few seconds)\n");
    printf("  --reset_mcu [-r] Reset the module (takes a few seconds)\n");
    printf("  --xmit [-x] Send payload to gateway (only valid for an endpoint)\n");
    printf("  --addr Specify the MAC address to be used with mail sent via --mail command to an endpoint\n");
    printf("  --smail Send mail message to endpoint (use --addr to specify its MAC address)\n");
    printf("    --smail 0 indicates a zero length mail is to be sent\n");
    printf("  --rmail Receive mail message in endpoint\n");
    printf("  --key [-k] Set the encryption key (specify -k 0 to remove existing key)\n");
    printf("  --sleep_block Block sleep \n");
    printf("  --sleep_unblock Unblock sleep \n");
    printf("  --count [-c] Return the # of msgs storged by a node (in EP its outgoing, GW - incoming)\n");
    printf("  --lost Return the # of msgs that were lost (GW: storage full, EP: ACK failure)\n");
    printf("  --msg [-m] Return the next available message storged in a module\n");
    printf("  --MSG [-M] Place module in receive continuous mode\n");
    printf("  --MSGM Place module in receive continuous mode (with return mail (of same size) sent to EP)\n");
    printf("  --time_set [-T] Sync module's time to current time in PC\n");
    printf("  --time_get Get module's time\n");
    printf("  --irq_flags_get get state of IRQ Flags Register\n");
    printf("  --irq_flags_clear clear bits in IRQ Flags Register\n");
    printf("\n");

    printf("Available properties:\n");
    printf("\n");
    for(i = 0; i < ENSEMBLE_PROP_MAX; i++ )
    {
        if( i == ENSEMBLE_PROP_MSG_TRAFFIC_KEY )       // key propery managed by its own command
        {
            continue;
        }
        printf("\t%d\t%s\n", i, property_map[i].name);
    }
    printf("\n");
}

    //
    //  switch settings for this tool
    //
    static struct option long_options[] =
    {
            {"baudrate",			required_argument, 0, 'b'},
            {"device",  			required_argument, 0, 'D'},
            {"version",             no_argument,       0, 'v'},
            {"set",                 required_argument, 0, 's'},
            {"get",                 no_argument,       0, 'g'},
            {"property",            no_argument,       0, 'p'},
            {"all",                 no_argument,       0, 'a'},
            {"mac_set",             required_argument, 0,   0},
            {"mac_get",             no_argument,       0,   0},
            {"help",     			no_argument,       0, 'h'},
            {"uuid",                no_argument,       0, 'u'},
            {"restore_defaults",    no_argument,       0, 'R'},
            {"irq_flags_clear",     required_argument, 0,   0},
            {"irq_flags_get",       no_argument,       0,   0},
            {"reset_mcu",  	    	no_argument,       0, 'r'},
            {"bootloader",          no_argument,       0, 'U'},
            {"delete_settings",     no_argument,       0, 'd'},
            {"save",                no_argument,       0, 'w'},
            {"xmit",                required_argument, 0, 'x'},
            {"smail",               required_argument, 0,   0},
            {"rmail",               no_argument,       0,   0},
            {"addr",                required_argument, 0,   0},
            {"sleep_block",         no_argument,       0,   0},
            {"sleep_unblock",       no_argument,       0,   0},
            {"antenna_set",         required_argument, 0,   0},
            {"antenna_get",         no_argument,       0,   0},
            {"key",                 required_argument, 0, 'k'},
            {"count",               no_argument,       0, 'c'},
            {"lost",                no_argument,       0,   0},
            {"msg",                 no_argument,       0, 'm'},
            {"MSG",                 no_argument,       0, 'M'},
            {"MSGM",                no_argument,       0,   0},
            {"time_set",            no_argument,       0, 'T'},
            {"time_get",            no_argument,       0,   0},
            {"dump",                no_argument,       0, 'z'},
            {0, 0, 0, 0}
    };

int main (int argc, char * const argv[])
{
    int c;
    int32_t i32_ret;    // Some function return int8_t's and some return int32_t's - know the difference!

    int i;

    // general buffer for propery value manipulation
    uint8_t property_value_buffer[(MAX_ENSEMBLE_PROPERTY_LENGTH * 2) + 1];     // note its size accomodates entering a key "string"
    uint8_t mac_address_buffer[(sizeof(uint64_t)* 2) + 1];                     // note its size accomodates entering a mac address "string"
    uint8_t payload_value_buffer[MAX_ENSEMBLE_TRANSFER_SIZE + 1];
    uint8_t payload_length;

    // other default parameters
    int baudrate = DEFAULT_BAUDRATE;
    char *dev_name = NULL;
    char *irq_flag_str = NULL;
    uint8_t reset_mcu = 0;
    uint8_t bootloader_mode = 0;
    uint8_t mac_set = SENTINEL;
    uint8_t mac_get = 0;
    uint8_t uuid = 0;
    uint8_t cmd_antenna_get = 0;
    uint8_t cmd_antenna_set = 0;
    uint8_t antenna_cfg = 0;
    uint8_t block_sleep = 0;
    uint8_t unblock_sleep = 0;
    uint8_t delete_settings = 0;
    uint8_t save_settings = 0;
    uint8_t restore_defaults = 0;
    uint8_t dump_settings = 0;
    uint8_t cmd_print_version = 0;
    uint8_t print_help = 0;
    uint8_t cmd_get_property = 0;
    uint8_t cmd_send_payload = 0;
    uint8_t cmd_send_mail = 0;
    uint8_t cmd_rcv_mail = 0;
    uint8_t cmd_set_macaddress = 0;
    uint8_t cmd_set_property = 0;
    uint8_t cmd_get_msg_cnt = 0;
    uint8_t cmd_get_lost_msg_cnt = 0;
    uint8_t cmd_get_msg = 0;
    uint8_t cmd_continuous_get_msg = 0;
    uint8_t cmd_dump_debug = 0;
    uint8_t cmd_set_key = 0;
    uint8_t cmd_set_time = 0;
    uint8_t cmd_get_time = 0;
    uint8_t cmd_return_mail = 0;
    uint8_t cmd_irq_flags_get = 0;
    uint8_t cmd_irq_flags_clear = 0;

    uint32_t    property_value = SENTINEL;          // user entered property value
    uint32_t    stored_msg_count = 0;               // # of msgs currently stored in a module
    uint32_t    lost_msg_count = 0;                 // count # of lost msgs (could not be stored - GW, or ACK failure - EP)
    ensemble_property_t property_id = ENSEMBLE_PROP_MAX;   // and the property ID

    uint8_t     rx_buffer[MAX_ENSEMBLE_TRANSFER_SIZE+1];  // Rx buffer for incoming msgs
    uint8_t     mail_buffer[MAX_ENSEMBLE_TRANSFER_SIZE+1]; // Mail buffer for outgoing mail msg
    uint8_t     dbg_buffer[ENSEMBLE_DBG_BUFFER_SIZE+1];  // Rx buffer for incoming msgs
    ensemble_msg_descriptor_t msg_descriptor;      // message descriptor structure (RSSI, time of Rx, MAC addr, etc)
    struct tm*  time_info;
    char timeString[80];
    time_t current_time;

    setbuf(stdout, NULL);                           // don't buffer internally - allows for better integration with Python Test Framework

    /* Use getopt to parse command line arguments */
    while (1)
    {
        /* getopt_long stores the option index here. */
        int option_index = 0;

        c = getopt_long(argc, argv, "b:D:vs:gp:ahuRrUdwcmMzx:k:T", long_options, &option_index);

        /* Detect the end of the options. */
        if (c == -1)
        {
            break;
        }

        switch (c)
        {
            case 0:
                /* If this option set a flag, do nothing else now. */
                if (long_options[option_index].flag != 0)
                    break;

                // get MAC mode (Nomac, LL, Micromode, etc)
                if(strcmp(long_options[option_index].name, "mac_get") == 0)
                {
                    mac_get = 1;
                }
                // set MAC mode
                else if(strcmp(long_options[option_index].name, "mac_set") == 0)
                {
                    mac_set = (uint8_t)strtol(optarg, NULL, 0);
                }
                // set antenna mode
                else if(strcmp(long_options[option_index].name, "antenna_set") == 0)
                {
                    antenna_cfg = (uint8_t)strtol(optarg, NULL, 0);
                    cmd_antenna_set = 1;
                }
                // get antenna mode
                else if(strcmp(long_options[option_index].name, "antenna_get") == 0)
                {
                    cmd_antenna_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "irq_flags_get") == 0)
                {
                    cmd_irq_flags_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "irq_flags_clear") == 0)
                {
                    irq_flag_str = optarg;
                    cmd_irq_flags_clear = 1;
                }
                else if(strcmp(long_options[option_index].name, "sleep_block") == 0)
                {
                	block_sleep = 1;
                }
                else if(strcmp(long_options[option_index].name, "sleep_unblock") == 0)
                {
                	unblock_sleep = 1;
                }
                else if(strcmp(long_options[option_index].name, "count") == 0)
                {
                	cmd_get_msg_cnt = 1;
                }
                else if(strcmp(long_options[option_index].name, "lost") == 0)
                {
                	cmd_get_lost_msg_cnt = 1;
                }
                else if(strcmp(long_options[option_index].name, "msg") == 0)
                {
                	cmd_get_msg = 1;
                }
                else if(strcmp(long_options[option_index].name, "MSG") == 0)
                {
                	cmd_continuous_get_msg = 1;
                }
                else if(strcmp(long_options[option_index].name, "MSGM") == 0)
                {
                	cmd_continuous_get_msg = 1;
                	cmd_return_mail = 1;
                }
                else if(strcmp(long_options[option_index].name, "time_set") == 0)
                {
                	cmd_set_time = 1;
                }
                else if(strcmp(long_options[option_index].name, "time_get") == 0)
                {
                	cmd_get_time = 1;
                }
                else if(strcmp(long_options[option_index].name, "dump") == 0)
                {
                	cmd_dump_debug = 1;
                }
                else if(strcmp(long_options[option_index].name, "key") == 0)
                {
                	cmd_set_key = 1;
                }
                else if(strcmp(long_options[option_index].name, "rmail") == 0)
                {
                	cmd_rcv_mail = 1;
                }
                else if(strcmp(long_options[option_index].name, "smail") == 0)
                {
                	cmd_send_mail = 1;
                    // the payload is just going to be an entered string (truncate size if too big)
                    i = strlen( optarg );
                    if( i > MAX_ENSEMBLE_PAYLOAD_LENGTH )
                    {
                        payload_length = MAX_ENSEMBLE_PAYLOAD_LENGTH;
                    }
                    // this is a special case to allow a zero length payload to be send to module
                    else if( (strlen( optarg ) == 1) && (optarg[0] == '0') )
                    {
                        payload_length = 0;
                    }
                    else
                    {
                        payload_length = i;
                    }
                    payload_value_buffer[payload_length] = 0;   // doing this so we can print it later (not needed for any other reason)
                    memcpy( payload_value_buffer, optarg, payload_length );
                }
                else if(strcmp(long_options[option_index].name, "addr") == 0)
                {
                	cmd_set_macaddress = 1;
                    strncpy( (char*)mac_address_buffer, optarg, (sizeof(uint64_t) * 2));
                    mac_address_buffer[sizeof(uint64_t) * 2] = 0;
                }
                else
                {
                    printf("option %s", long_options[option_index].name);
                }
                break;

            // dump all property values if MAC is in ensemble mode
            case 'a':
                dump_settings = 1;
                break;

            //  set the baud rate
            case 'b':
                baudrate = (int)strtol(optarg, NULL, 0);
                break;

            // delete all configuration settings
            case 'd':
                delete_settings = 1;
                break;

            // set the com port/device to use to communicate with the module
            case 'D':
                dev_name = optarg;
                break;

            // help text dump
            case 'h':
                print_help = 1;
                break;

            // help text dump
            case 'u':
                uuid = 1;
                break;

            // restart the module
            case 'r':
                reset_mcu = 1;
                break;

            // restore settings to defaults
            case 'R':
                restore_defaults = 1;
                break;

            // save settings to flash
            case 'w':
                save_settings = 1;
                break;

            // dump version info about tool
            case 'v':
                cmd_print_version = 1;
                break;

            // get the current msg count number stored in the module
            case 'c':
                cmd_get_msg_cnt = 1;
                break;
            // get the next msg in the module
            case 'm':
                cmd_get_msg = 1;
                break;
            // continuous get msg
            case 'M':
                cmd_continuous_get_msg = 1;
                break;
            // set time (sync to current PC time)
            case 'T':
                cmd_set_time = 1;
                break;
            // dump debug data
            case 'z':
                cmd_dump_debug = 1;
                break;
            // set encryption key
            case 'k':
                cmd_set_key = 1;
                                    // copy in the key, process it later (note copy length is string length not binary key length)
                strncpy( (char*)property_value_buffer, optarg, (ENSEMBLE_AES_KEY_LENGTH * 2) + 1);
                break;

            // send payload (to gateway)
            case 'x':
                cmd_send_payload = 1;
                // the payload is just going to be an entered string (truncate size if too big)
                i = strlen( optarg );
                if( i > MAX_ENSEMBLE_PAYLOAD_LENGTH )
                {
                    payload_length = MAX_ENSEMBLE_PAYLOAD_LENGTH;
                }
                else
                {
                    payload_length = i;
                }
                payload_value_buffer[payload_length] = 0;   // doing this so we can print it later (not needed for any other reason)
                memcpy( payload_value_buffer, optarg, payload_length );
                break;

            // reset into bootstrap mode
            case 'U':
                bootloader_mode = 1;
                break;

            // set property request
            case 's':
                cmd_set_property = 1;
                // we are going to cheat for now and not handle the case where this is a key
                property_value = (uint32_t)strtol(optarg, NULL, 0);
                // put the value in an outgoing buffer. key property is handled by a different switch
                // if in the future there are properties of non- uint32 sizes, this has to be updated
                *((uint32_t *)property_value_buffer) = property_value;
                break;

            // get property request
            case 'g':
                cmd_get_property = 1;
                break;

            // set property request
            case 'p':
                property_id = (uint32_t)strtol(optarg, NULL, 0);
                if( property_id >= ENSEMBLE_PROP_MAX )
                {
                    printf("Invalid property number (%d) specified!\n", property_id);
                    usage();
                    exit(EXIT_SUCCESS);
                }
                break;

            // option is bogus
            default:
                printf("Invalid option entered\n");
                usage();
                exit(1);
                break;
        }
    }
    //  handle each command one at a time

    // print dumps help text and quits
    if (print_help)
    {
        usage();
        exit(EXIT_SUCCESS);
    }

    // dump simple tool version info
    if(cmd_print_version)
    {
        printf("Executable Version = %d.%d.%d\n", VERSION_MAJOR, VERSION_MINOR, VERSION_TAG);
    }

    // open the com port at the correct baud rate
    ll_tty_open(dev_name, baudrate);

    // only print this stuff if nothing else was entered as a switch (beside com port setting)
    if( (argc == 1 ) || (( argc == 3 ) && (dev_name != NULL) ))
    {
        print_ll_firmware_type();
        print_ll_hardware_type();
        print_ll_version();
        print_uuid();
    }

    // get the MAC mode if it was asked for
    if(mac_get)
    {
        ll_mac_type_t mac_mode = MAC_INVALID;
        i32_ret = ll_mac_mode_get(&mac_mode);
        print_ll_ifc_error("ll_mac_mode_get", i32_ret);
        printf("MAC Mode = %d\n", mac_mode);
    }
    // set the MAC mode if requested
    if(mac_set != SENTINEL)
    {
        i32_ret = ll_mac_mode_set(mac_set);
        print_ll_ifc_error("ll_mac_mode_set", i32_ret);
        printf("set MAC mode to %d\n", mac_set);
    }
    // get the antenna mode if it was asked for
    if (cmd_antenna_get)
    {
        uint8_t ant;
        i32_ret = ll_antenna_get(&ant);
        print_ll_ifc_error("ll_antenna_get", i32_ret);
        if (i32_ret >= 0)
        {
            printf("Antenna Configuration = %d\n", ant);
        }
    }
    // set the antenna mode if requested
    if (cmd_antenna_set)
    {
        i32_ret = ll_antenna_set(antenna_cfg);
        print_ll_ifc_error("ll_receive_mode_set", i32_ret);
    }

    if(block_sleep)
    {
    	i32_ret = ll_sleep_block();
    	print_ll_ifc_error("ll_sleep_block", i32_ret);
    }

    if(unblock_sleep)
    {
    	i32_ret = ll_sleep_unblock();
    	print_ll_ifc_error("ll_sleep_unblock", i32_ret);
    }

    if((cmd_irq_flags_get) || (cmd_irq_flags_clear))
    {
        uint32_t irq_flags_to_clear = 0;
        uint32_t irq_flags_read = 0;
        uint8_t len = 4;
        int8_t i8_ret;

        if (cmd_irq_flags_clear)
        {
            len = strlen(irq_flag_str)/2;
            buffer_to_hex(irq_flag_str);
            irq_flags_to_clear = ((uint32_t)irq_flag_str[3]);
            irq_flags_to_clear |= ((uint32_t)irq_flag_str[2]) << 8;
            irq_flags_to_clear |= ((uint32_t)irq_flag_str[1]) << 16;
            irq_flags_to_clear |= ((uint32_t)irq_flag_str[0]) << 24;
        }

        if (len != 4)
        {
            printf("Argument must be 32-bit hex\n");
        }
        else
        {
            // Check what flags are set, and clear some if we want
            i8_ret = ll_irq_flags(irq_flags_to_clear, &irq_flags_read);
            if (i8_ret<0)
            {
                print_ll_ifc_error("ll_irq_flags", i8_ret);
            }
            else
            {
                if(cmd_irq_flags_get)
                {
                    printf("irq_flags = 0x%08X\n", irq_flags_read);
                    print_irq_flags_text(irq_flags_read);
                }
            }
        }
    }

    // dump the uuid if requested
    if (uuid)
    {
        print_uuid();
    }

    // delete the settings if requested
    if (delete_settings)
    {
        i32_ret = ll_settings_delete();
        print_ll_ifc_error("ll_settings_delete", i32_ret);
    }

    // save settings if requested
    if (save_settings)
    {
        printf("Saving module settings\n");
        i32_ret = ll_settings_store();
        print_ll_ifc_error("ll_store_settings", i32_ret);
    }

    // restore defaults if requested
    if (restore_defaults)
    {
        i32_ret = ll_restore_defaults();
        print_ll_ifc_error("ll_restore_defaults", i32_ret);
    }

    // reset module if requested
    if (reset_mcu)
    {
        printf("Resetting module\n");
        i32_ret = ll_reset_mcu();
        print_ll_ifc_error("ll_reset_mcu", i32_ret);
    }

    // reset into bootloader if requested
    if (bootloader_mode)
    {
        printf("Putting module in bootloader mode\n");
        i32_ret = ll_bootloader_mode();
        print_ll_ifc_error("ll_bootloader_mode", i32_ret);
    }

    // get a property
    if(cmd_get_property)
    {
                        // must have specified a property but key is handled in its own switch setting
        if( (property_id >= ENSEMBLE_PROP_MAX) || (property_id == ENSEMBLE_PROP_MSG_TRAFFIC_KEY))
        {
            printf("Must specify a valid property ID [-p] to perform the get!\n");
            usage();
            exit(EXIT_SUCCESS);
        }
        // property was specified if we get here, so go try to get it
        i32_ret = ll_get_ensemble_config_property( property_id, property_value_buffer );
        print_ll_ifc_error("ll_get_ensemble_config_property", i32_ret);

        if( i32_ret > 0 )
        {
            // we are only going to handle 32 bit integer reads right now
            if( ll_ensemble_property_size(property_id) == sizeof(uint32_t) )
            {
                printf("Ensemble Property %d (%s) = %d\n", property_id, propertyNameToStr(property_id), *((uint32_t *)(property_value_buffer)) );
            }
            {
                ; // future if properties are added of other data types
            }
        }
    }

    // set a property
    if(cmd_set_property)
    {
                        // must have specified a property but key is handled in its own switch setting
        if( (property_id >= ENSEMBLE_PROP_MAX) || (property_id == ENSEMBLE_PROP_MSG_TRAFFIC_KEY))
        {
            printf("Must specify a valid property ID [-p] to perform the set!\n");
            usage();
            exit(EXIT_SUCCESS);
        }
        // property was specified if we get here, so go try to get it
        i32_ret = ll_set_ensemble_config_property( property_id, property_value_buffer );
        print_ll_ifc_error("ll_set_ensemble_config_property", i32_ret);

        if( i32_ret >= 0 )
        {
            printf("Ensemble Property %d (%s)  was set to: %d\n", property_id, propertyNameToStr(property_id), *((uint32_t *)(property_value_buffer)) );
        }
    }

    // set a property
    if(cmd_set_key)
    {
        int key_cleared = 0;

                        // must key length must be 2x size binary size of key
        if( strlen( (char*)property_value_buffer ) != ENSEMBLE_AES_KEY_LENGTH * 2 )
        {
            // a '0' specified as a key means clear the current key
            if( (strlen((char*)property_value_buffer) == 1) && (*(char*)property_value_buffer == '0') )
            {
                memset( property_value_buffer, 0, ENSEMBLE_AES_KEY_LENGTH );
                key_cleared = 1;
            }
            else
            {
                printf("Must specify a %d digit key!\n", ENSEMBLE_AES_KEY_LENGTH * 2);
                usage();
                exit(EXIT_SUCCESS);
            }
        }
        else
        {
                        // key completely specified, convert to hex (in place)
                        // length when packed as binary will be the key length
            if (buffer_to_hex( (char*)property_value_buffer ) != ENSEMBLE_AES_KEY_LENGTH )
            {
                printf("Invalid entry digits mus be HEX (0-9) or (A-F)!\n");
                usage();
                exit(EXIT_SUCCESS);
            }
        }
        i32_ret = ll_set_ensemble_config_property( ENSEMBLE_PROP_MSG_TRAFFIC_KEY, property_value_buffer );
        print_ll_ifc_error("ll_set_ensemble_config_property (key)", i32_ret);
        if( i32_ret >= 0 )
        {
            if( key_cleared )
            {
                printf("Ensemble Encryption key set to default value.\n");
            }
            else
            {
                printf("Ensemble Encryption key was set to: 0x");
                for( i = 0; i < ENSEMBLE_AES_KEY_LENGTH; i++ )
                {
                    printf("%x", property_value_buffer[i]);
                }
                printf("\n");
            }
        }
    }

    // set a property
    if(cmd_set_macaddress)
    {
        uint8_t work_buffer[(sizeof(uint64_t) * 2) + 1];
        uint8_t addr_size;

        addr_size = strlen( (char*)mac_address_buffer );    // we use this a lot

        // range check (string size)
        if( addr_size > (sizeof(uint64_t) * 2) )
        {
            printf("MAC address length too long (max of %u digits)\n", (uint32_t)(sizeof(uint64_t) * 2));
            usage();
            exit(EXIT_SUCCESS);
        }
        // prefix what's before the address (if its not 16 digits with '0's
        memset( work_buffer, '0', sizeof(uint64_t) * 2);
        // overlay value in right spot
        memcpy( &work_buffer[(sizeof(uint64_t) * 2) - addr_size], mac_address_buffer, addr_size);
        // copy it back to original buffer
        memcpy( mac_address_buffer, work_buffer, sizeof(uint64_t) * 2);
        // now convert to binary from hex digits
        if (buffer_to_hex( (char*)mac_address_buffer ) != sizeof(uint64_t) )
        {
            printf("Invalid entry digits mus be HEX (0-9) or (A-F)!\n");
            usage();
            exit(EXIT_SUCCESS);
        }
    }

    // dump all the properties
    if(dump_settings)
    {
        printf("\n");
        printf("List of all Ensemble property values: \n");
        for( i = 0; i < ENSEMBLE_PROP_MAX; i++ )
        {
            if( i == ENSEMBLE_PROP_MSG_TRAFFIC_KEY )       // key property not accessible by "get"
            {
                continue;
            }
            i32_ret = ll_get_ensemble_config_property( i, property_value_buffer );
            print_ll_ifc_error("ll_get_ensemble_config_property", i32_ret);    // dump error and quit if it starts bombing out
            if( i32_ret < 0 )
                break;
                                                                                // otherwise we print them
            if( ll_ensemble_property_size(i) == sizeof(uint32_t) )
            {
                printf("Ensemble Property %d (%s) = %d\n", i, propertyNameToStr(i), *((uint32_t *)(property_value_buffer)) );
            }
            {
                ; // future if properties are added of other data types
            }
        }
    }

    // send msg to gateway
    if(cmd_send_payload)
    {
        // send the payload...(actually enqueue for transmission)
        i32_ret = ll_send_payload_to_gw( payload_value_buffer, payload_length );
        print_ll_ifc_error("ll_send_payload_to_gw", i32_ret);
        if( i32_ret >= 0 )
        {
            printf("Payload %s of size %d has been queued for Tx\n", payload_value_buffer, payload_length  );
        }
    }
    // send mail to gateway (for transmission to EP of specified MAC address)
    if(cmd_send_mail)
    {
        // send the payload...(actually enqueue for transmission)
        // note a special case (0 length payload is permitted)

        // if mac wasn't specified its set to broadcast
        if( cmd_set_macaddress == 0 )
        {
            memset( mac_address_buffer, 0, sizeof(uint64_t) );
        }
        i32_ret = ll_send_mail_to_ep( (uint64_t *)mac_address_buffer, payload_value_buffer, payload_length );
        print_ll_ifc_error("ll_send_mail_to_ep", i32_ret);
        if( i32_ret >= 0 )
        {
            printf("Mail (content = %s) size=%d has been queued for MAC=0x", payload_value_buffer, payload_length  );
            for( i = 0; i < (int)sizeof(uint64_t); i++ )
            {
                printf("%02x", mac_address_buffer[i]);
            }
            printf("\n");
        }
    }
    // get # of msgs stored in module
    if(cmd_get_msg_cnt)
    {
        // send the payload...(actually enqueue for transmission)
        i32_ret = ll_get_ensemble_stored_msg_count( &stored_msg_count );
        print_ll_ifc_error("ll_get_ensemble_stored_msg_count", i32_ret);
        if( i32_ret >= 0 )
        {
            printf("Module Rx'ed msg count = %d\n", stored_msg_count  );
        }
    }
    // get # of lost msgs stored in module or in EP msgs that ACK failed
    if(cmd_get_lost_msg_cnt)
    {
        // send the payload...(actually enqueue for transmission)
        i32_ret = ll_get_ensemble_lost_msg_count( &lost_msg_count, 1 );        // reset the count
        print_ll_ifc_error("ll_get_ensemble_lost_msg_count", i32_ret);
        if( i32_ret >= 0 )
        {
            printf("Module lost msg count = %d\n", lost_msg_count  );
        }
    }
    // get next message stored in the module
    if(cmd_get_msg)
    {
        uint32_t    mac_upper;
        uint32_t    mac_lower;

        // get the msg (if one exists)
        i32_ret = ll_get_ensemble_get_stored_msg(&msg_descriptor, rx_buffer, MAX_ENSEMBLE_TRANSFER_SIZE + 1);
        print_ll_ifc_error("ll_get_ensemble_get_stored_msg", i32_ret);
        if( i32_ret >= 0 )
        {
            mac_upper = (uint32_t)(msg_descriptor.mui >> 32);
            mac_lower = (uint32_t)(msg_descriptor.mui & 0xFFFFFFFF);
            printf("MAC Addr: 0x%08x", mac_upper);
            printf("%08x ", mac_lower);
                            // if time not set...just say so
            if( msg_descriptor.utc_time == 0 )
            {
                printf("Msg Time: <not set>: ");
            }
            else
            {
                current_time = msg_descriptor.utc_time;
                time_info = localtime( &current_time );
                strftime(timeString,TIME_BUFFER_SIZE,"%x %X", time_info);
                printf("Msg Time: %s: ", timeString );
            }
                // RSSI
            printf("RSSI = %d dBm ", msg_descriptor.rssi );
                // msg characteristics field
            if( msg_descriptor.msginfo & ENSEMBLE_MSG_ROUTED_VIA_REPEATER )
            {
                printf("Via Repeater ");
            }
            else
            {
                printf("From EP ");
            }
                // content dump
            printf("Size: %d\n\n", i32_ret );
            for (i = 0; i < i32_ret; i++)
            {
                if( (i % 16) == 0 )
                {
                    printf("%02x: ", i);
                }
                printf("%02x ", rx_buffer[i]);
                if( (i % 16) == 15 )
                {
                    printf("\n");
                }
            }
            printf("\n");
        }
    }
    // get next mail message stored in the endpoint
    if(cmd_rcv_mail)
    {
        // get the msg (if one exists)
        i32_ret = ll_get_ensemble_get_mail_msg(rx_buffer, MAX_ENSEMBLE_TRANSFER_SIZE + 1);
        print_ll_ifc_error("ll_get_ensemble_get_mail_msg", i32_ret);

        if( i32_ret >= 0 )
        {
            // null terminate the string
            rx_buffer[i32_ret] = 0;
            printf("Mail Rcvd: Size = %d Content = %s\n", i32_ret, rx_buffer );
        }
    }
    // get / dump debug information
    if(cmd_dump_debug)
    {
        // get the msg (if one exists)
        i32_ret = ll_get_ensemble_get_debuginfo(dbg_buffer, ENSEMBLE_DBG_BUFFER_SIZE + 1);
        print_ll_ifc_error("ll_get_ensemble_get_debuginfo", i32_ret);
        if( i32_ret >= 0 )
        {
            printf("DEBUG info (size %d)\n", i32_ret );
            for (i = 0; i < i32_ret; i++)
            {
                if( (i % 16) == 0 )
                {
                    printf("%02x: ", i);
                }
                printf("%02x ", dbg_buffer[i]);
                if( (i % 16) == 15 )
                {
                    printf("\n");
                }
            }
            printf("\n");
        }
    }

    // get next message stored in the module but continue here forever (^C) gets you out
    if(cmd_continuous_get_msg)
    {
        int rcvd_msg_number = 0;
        uint32_t    mac_upper;
        uint32_t    mac_lower;
        unsigned    int i;
        int         msg_size;

        printf("Starting continuous receive...");
        if( cmd_return_mail )
        {
            printf("With Immediate Mail LOOPBACK");
        }
        printf("\n");
        while (1)
        {
            MS_SLEEP(100);
            // continue to query interface if msg exist
            i32_ret = ll_get_ensemble_stored_msg_count( &stored_msg_count );
            // no point in spamming if there's an error
            print_ll_ifc_error("ll_get_ensemble_stored_msg_count", i32_ret);
            if( i32_ret < 0 )
                break;

            if(  stored_msg_count > 0 )     // means @least one msg is available
            {
                i32_ret = ll_get_ensemble_get_stored_msg(&msg_descriptor, rx_buffer, MAX_ENSEMBLE_TRANSFER_SIZE + 1);
                                    // this would be strange...have msgs, but none is really returned
                print_ll_ifc_error("ll_get_ensemble_get_stored_msg", i32_ret);
                                    // dump the msg just like the other msg dump option does...
                if( i32_ret > 0 )
                {
                    rx_buffer[i32_ret] = 0;     // null terminate the string we use it later
                    msg_size = i32_ret;

                                    // if mail requested loop back a mail msgs to the node that sent this
                                    // msg content will be the message in upper-case
                    if( cmd_return_mail > 0 )
                    {
                                    // copy into mail buffer and make upper case
                        memcpy( mail_buffer, rx_buffer, i32_ret );
                        mail_buffer[i32_ret] = 0;
                                    // old school uppercase conversion
                        for(i = 0; i < strlen((char *)mail_buffer); i++ )
                        {
                            mail_buffer[i] = toupper(mail_buffer[i]);
                        }
                                    // send in the mail
                        i32_ret = ll_send_mail_to_ep( &msg_descriptor.mui, mail_buffer, i32_ret );
                        print_ll_ifc_error("ll_send_mail_to_ep", i32_ret);
                    }
                    mac_upper = (uint32_t)(msg_descriptor.mui >> 32);
                    mac_lower = (uint32_t)(msg_descriptor.mui & 0xFFFFFFFF);
                    rcvd_msg_number++;
                    printf("MAC Addr = 0x%08x", mac_upper);
                    printf("%08x ", mac_lower);
                    printf("Msg # = %03d Size = %03d ", rcvd_msg_number, msg_size );
                            // if time not set...just say so
                    if( msg_descriptor.utc_time == 0 )
                    {
                        printf("Msg Time = <Unknown> ");
                    }
                    else
                    {
                        current_time = msg_descriptor.utc_time;
                        time_info = localtime( &current_time );
                        strftime(timeString,TIME_BUFFER_SIZE,"%x %X", time_info);
                        printf("Msg Time = %s: ", timeString );
                    }
                        // RSSI
                    printf("RSSI = %d dBm ", msg_descriptor.rssi );
                        // msg characteristics field
                    if( msg_descriptor.msginfo & ENSEMBLE_MSG_ROUTED_VIA_REPEATER )
                    {
                        printf("Via Repeater ");
                    }
                    else
                    {
                        printf("From EP ");
                    }
                    printf("Content = %s\n", rx_buffer);
                }
            }
        }
    }
    // set the current module's time
    if(cmd_set_time)
    {
        // get current PC time (in UTC) and send to module
        // the issue is that its not going to be absolutely precise (propagation delay), so we'll do our best to set right at the second' transition
        uint64_t    module_time;

        current_time = time(NULL);
        while( time(NULL) < current_time );     // tight poll loop waiting for a 1 sec rollover
                                                // best we can to be exact for time set
        current_time = time(NULL);              // doing it this way to handle conversion to uint64 and fact we can't trust rollover is more than 1 sec
        module_time = current_time;

        i32_ret = ll_set_utc_time(&module_time);
        print_ll_ifc_error("ll_set_utc_time", i32_ret);
        if( i32_ret >= 0 )
        {
           time_info = localtime( &current_time );
            strftime(timeString,TIME_BUFFER_SIZE,"%x %X", time_info);
            printf("The module time is now = %s\n", timeString );
        }
    }
    // get the current module time
    if(cmd_get_time)
    {
        // get current PC time (in UTC) and send to module
        // the issue is that its not going to be absolutely precise (propagation delay), so we'll do our best to set right at the second' transition

        uint64_t    module_time;
        struct tm*  time_info;
        char timeString[80];

        i32_ret = ll_get_utc_time(&module_time);
        print_ll_ifc_error("ll_get_utc_time", i32_ret);
        if( i32_ret >= 0 )
        {
                        // convert to a time_t (if required)
                        // then dump the time (if its non-zero). A zero value means the module time was never set
            current_time = module_time;
            if( current_time == 0 )
            {
                printf("The module time has never been set!\n");
            }
            else
            {
                time_info = localtime( &current_time );
                strftime(timeString,TIME_BUFFER_SIZE,"%x %X", time_info);
                printf("The current module time is = %s\n", timeString );
            }
        }
    }
    return 0;
}
