#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <sys/timeb.h>

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#if defined(WIN32) || defined (__MINGW32__)
#include <windows.h>
#endif

#include "ifc_struct_defs.h"
#include "ll_ifc_consts.h"
#include "ll_ifc_transport_pc.h"
#include "ll_ifc_symphony.h"

#define NEW_HOST_IFC

#if defined(WIN32) || defined (__MINGW32__)
#define SLEEP(x)        Sleep(x * 1000)
#define MS_SLEEP(x)     Sleep(x)
#else
#define SLEEP(x)        sleep(x)
#define MS_SLEEP(x)     usleep(x * 1000)
#endif

#define DEFAULT_CHANNEL            (0)
#define DEFAULT_BANDWIDTH        (3)
#define SENTINEL                (255)

// Helper macros to convert little-endian byte stream to native types
#define _HOST_IFC_LE_U32(x,o,y) ((((uint32_t)(((uint8_t*)&x)[o+y]))&0xFF) << (8*(y)))
#define HOST_IFC_LE_U32(x,o) (_HOST_IFC_LE_U32(x,o,0) + _HOST_IFC_LE_U32(x,o,1) + _HOST_IFC_LE_U32(x,o,2) + _HOST_IFC_LE_U32(x,o,3))

#define _HOST_IFC_LE_I16(x,o,y) ((((int16_t)(((uint8_t*)&x)[o+y]))&0xFF) << (8*(y)))
#define HOST_IFC_LE_I16(x,o) (_HOST_IFC_LE_I16(x,o,0) + _HOST_IFC_LE_I16(x,o,1))

// Build Date
#define BUILD_DATE (__BUILD_DATE);

// Build Number
#define VERSION_MAJOR                (1) // 0-255
#define VERSION_MINOR                (0) // 0-255
#define VERSION_TAG     (__BUILD_NUMBER) // 0-65535


typedef enum
{
    MODE_SLEEP,
    MODE_SEND_ACK,
    MODE_SEND_UNACK,
    MODE_NONE
} module_mode_t;

typedef struct _echo_test
{
    short cnt;
    char payload[64];
} echo_test_t;

static int g_print_help_exit = 0;

static void print_ll_ifc_error(char *label, int32_t ret_val)
{
    if (ret_val < 0)
    {
        fprintf(stderr, "ERROR(%s): Host interface - ", label);

        // Map Error code to NACK code
        if (ret_val >= -99 && ret_val <= -1)
        {
            ret_val = 0 - ret_val;
        }
        switch(ret_val)
        {
            case LL_IFC_NACK_CMD_NOT_SUPPORTED:
                fprintf(stderr, "NACK received - Command not supported");
                break;
            case LL_IFC_NACK_INCORRECT_CHKSUM:
                fprintf(stderr, "NACK received - Incorrect Checksum");
                break;
            case LL_IFC_NACK_PAYLOAD_LEN_OOR:
                fprintf(stderr, "NACK received - Payload length out of range");
                break;
            case LL_IFC_NACK_PAYLOAD_OOR:
                fprintf(stderr, "NACK received - Payload out of range");
                break;
            case LL_IFC_NACK_BOOTUP_IN_PROGRESS:
                fprintf(stderr, "NACK received - Not allowed, bootup in progress");
                break;
            case LL_IFC_NACK_BUSY_TRY_AGAIN:
                fprintf(stderr, "NACK received - Busy try again");
                break;
            case LL_IFC_NACK_APP_TOKEN_REG:
                fprintf(stderr, "NACK received - Application Token not registered");
                break;
            case LL_IFC_NACK_PAYLOAD_LEN_EXCEEDED:
                fprintf(stderr, "NACK received - Payload length greater than maximum");
                break;
            case LL_IFC_NACK_NOT_IN_MAILBOX_MODE:
                fprintf(stderr, "NACK received - Module is not in DOWNLINK_MAILBOX mode");
                break;
            case LL_IFC_NACK_OTHER:
                fprintf(stderr, "NACK received - Other");
                break;
            case LL_IFC_ERROR_INCORRECT_PARAMETER:
                fprintf(stderr, "Invalid Parameter");
                break;
            case -103:
                fprintf(stderr, "Message Number mismatch");
                break;
            case -104:
                fprintf(stderr, "Checksum mismatch");
                break;
            case -105:
                fprintf(stderr, "Command mismatch");
                break;
            case -106:
                fprintf(stderr, "Timed out");
                break;
            case -107:
                fprintf(stderr, "Payload larger than buffer provided");
                break;
            default:
                break;
        }
        fprintf(stderr, "\n");
    }
}

static void print_ll_version(void)
{
    ll_version_t ver;
    int8_t i8_ret;

    i8_ret = ll_version_get(&ver);
    print_ll_ifc_error("ll_version_get", i8_ret);
    if (i8_ret == VERSION_LEN)
    {
        printf("Link Labs Version: %d.%d.%d\n", ver.major, ver.minor, ver.tag);
    }
}

static void print_ll_ifc_version(void)
{
    ll_version_t ver;
    int8_t i8_ret;

    i8_ret = ll_interface_version_get(&ver);
    print_ll_ifc_error("ll_interface_version_get", i8_ret);
    if (i8_ret == VERSION_LEN)
    {
        printf("Link Labs interface version: %d.%d.%d\n", ver.major, ver.minor, ver.tag);

        if (!(ver.major == IFC_VERSION_MAJOR && ver.minor == IFC_VERSION_MINOR && ver.tag == IFC_VERSION_TAG))
        {
            fprintf(stderr, "host interface version mismatch!\n");
        }
    }
}

static void print_uuid(void)
{
    uint64_t uuid = 0;
    int8_t ret;

    ret = ll_unique_id_get(&uuid);
    uuid &= 0xFFFFFFFFF;
    print_ll_ifc_error("ll_unique_id_get", ret);
    if (ret >= 0)
    {
        printf("Link Labs UUID: $301$0-0-0-%01x", (uint32_t)(uuid >> 32));
        printf("%08x\n", (uint32_t)uuid);
    }
}

static ll_firmware_type_t s_fw_type;

static void print_ll_firmware_type(void)
{
    int8_t ret;

    ret = ll_firmware_type_get(&s_fw_type);
    print_ll_ifc_error("ll_firmware_type", ret);

    if (ret == FIRMWARE_TYPE_LEN)
    {
        printf("Link Labs Firmware Type: %04d.%04d\n", s_fw_type.cpu_code, s_fw_type.functionality_code);
    }
}

static void print_ll_hardware_type(void)
{
    ll_hardware_type_t t;
    int8_t ret;

    ret = ll_hardware_type_get(&t);
    print_ll_ifc_error("ll_hardware_type", ret);

    if (ret == 1)
    {
        switch(t)
        {
            case UNAVAILABLE:
                break;
            case LLRLP20_V2:
                printf("Link Labs Hardware Type: LLRLP20 v2\n");
                break;
            case LLRXR26_V2:
                printf("Link Labs Hardware Type: LLRXR26 v2\n");
                break;
            case LLRLP20_V3:
                printf("Link Labs Hardware Type: LLRLP20 v3\n");
                break;
            case LLRXR26_V3:
                printf("Link Labs Hardware Type: LLRXR26 v3\n");
                break;
            default:
                break;
        }
    }
}

static void print_band_cfg(llabs_dl_band_cfg_t* p)
{
    printf("  band_edge_lower  : %d [Hz]\n", p->band_edge_lower);
    printf("  band_edge_upper  : %d [Hz]\n", p->band_edge_upper);
    printf("  band_edge_guard  : %d [Hz]\n", p->band_edge_guard);
    printf("  chan_step_size   : %d [Channels]\n", p->chan_step_size);
    printf("  chan_step_offset : %d [Channels]\n", p->chan_step_offset);
}

static void print_irq_flags_text(uint32_t flags)
{
    // Most significant bit to least significant bit
    if(IRQ_FLAGS_ASSERT & flags)
    {
        printf("[IRQ_FLAGS_ASSERT]");
    }
    if(IRQ_FLAGS_APP_TOKEN_ERROR & flags)
    {
        printf("[IRQ_FLAGS_APP_TOKEN_ERROR]");
    }
    if(IRQ_FLAGS_CRYPTO_ERROR & flags)
    {
        printf("[IRQ_FLAGS_CRYPTO_ERROR]");
    }
    if(IRQ_FLAGS_DOWNLINK_REQUEST_ACK & flags)
    {
        printf("[IRQ_FLAGS_DOWNLINK_REQUEST_ACK]");
    }
    if(IRQ_FLAGS_APP_TOKEN_CONFIRMED & flags)
    {
        printf("[IRQ_FLAGS_APP_TOKEN_CONFIRMED]");
    }
    if(IRQ_FLAGS_CRYPTO_ESTABLISHED & flags)
    {
        printf("[IRQ_FLAGS_CRYPTO_ESTABLISHED]");
    }
    if(IRQ_FLAGS_DISCONNECTED & flags)
    {
        printf("[IRQ_FLAGS_DISCONNECTED]");
    }
    if(IRQ_FLAGS_CONNECTED & flags)
    {
        printf("[IRQ_FLAGS_CONNECTED]");
    }
    if(IRQ_FLAGS_RX_DONE & flags)
    {
        printf("[IRQ_FLAGS_RX_DONE]");
    }
    if(IRQ_FLAGS_TX_ERROR & flags)
    {
        printf("[IRQ_FLAGS_TX_ERROR]");
    }
    if(IRQ_FLAGS_TX_DONE & flags)
    {
        printf("[IRQ_FLAGS_TX_DONE]");
    }
    if(IRQ_FLAGS_RESET & flags)
    {
        printf("[IRQ_FLAGS_RESET]");
    }
    if(IRQ_FLAGS_WDOG_RESET & flags)
    {
        printf("[IRQ_FLAGS_WDOG_RESET]");
    }
    if(flags != 0)
    {
        printf("\n");
    }
}

// Returns 0 if good, negative if bad
// Accepts a comma seperated list of 5 values
static int32_t parse_band_cfg(char* arg_str, llabs_dl_band_cfg_t* p)
{
    char *token;
    uint32_t arg_count = 0;

    token = strtok(arg_str, ",");
    if (token == NULL)
    {
        return(-1);
    }
    arg_count++;

    while (token != NULL)
    {
        // printf( "%d --> %s\n", arg_count, token );
        switch (arg_count)
        {

            case 1:
                p->band_edge_lower = (uint32_t)strtol(token, NULL, 0);
                break;
            case 2:
                p->band_edge_upper = (uint32_t)strtol(token, NULL, 0);
                break;
            case 3:
                p->band_edge_guard = (uint32_t)strtol(token, NULL, 0);
                break;
            case 4:
                p->chan_step_size = (uint8_t)strtol(token, NULL, 0);
                break;
            case 5:
                p->chan_step_offset = (uint8_t)strtol(token, NULL, 0);
                break;
            default:
                {
                    return(-1);
                }
                break;
        }
        token = strtok(NULL, ",");
        arg_count++;
    }

    if (arg_count != 6)
    {
        return(-1);
    }

    printf("Setting Downlink Band Configuration to:\n");
    print_band_cfg(p);

    return(0);
}

static int32_t buffer_to_hex(char* buf)
{
    // only valid characters are 0:9,a:f,a:f
    uint16_t i;
    uint16_t num_chars = strlen(buf);
    if (num_chars&0x01)
    {
        fprintf(stderr, "hex payload must contain even number of characters\n");
        fprintf(stderr, "failed to send packet\n");
        exit(EXIT_FAILURE);
    }

    for(i=0; i<num_chars; i++)
    {
        char c = toupper(buf[i]);
        uint8_t x;

        // printf("\t\t%x\n", c);

        // 0:9
        if (isdigit(c))
        {
            x = c-'0';
        }
        else if ((c >= 'A') && (c <= 'F'))
        {
            x = c-'A'+10;
        }
        else
        {
            fprintf(stderr, "hex payload must only contain characters: 0-9, a-f, A-F\n");
            fprintf(stderr, "failed to send packet\n");
            exit(EXIT_FAILURE);
        }

        // pack the buffer
        if (i & 0x01)
        {
            // printf("low  [%d] = %x\n", (i-1)/2, x);
            buf[(i-1)/2] |= x;
        }
        else
        {
            // printf("high [%d] = %x\n", i/2, x<<4);
            buf[i/2] = x<<4;
        }
    }

    // printf("num_chars = %d\n", num_chars);
    buf[num_chars/2] = '\0';
    return strlen(buf);
}

static void usage(void)
{
    printf("Usage:\n");
    printf("  --help [-h] print this help message\n");
    printf("  --device [-D] choose tty device, default: "LL_TTY_DEFAULT_DEVICE"\n");
    printf("  --baudrate [-b] configure baudrate of the tty device, default: %d\n", LL_TTY_DEFAULT_BAUDRATE);
    printf("  --version [-v] returns the version of this executable\n");
    printf("\n");
    printf("CONFIGURATION\n");
    printf("  --net_token_set Set the network id\n");
    printf("  --app_token_set Set the application token\n");
    printf("  --receive_set Set downlink mode [0:off, 1:always on]\n");
    printf("  --qos_set Request QoS level [0-15]\n");
    printf("  --config_get Get the module's configuration for Symphony\n");
    printf("  --antenna_set Set the antenna configuration \n");
    printf("  --antenna_get Get the antenna configuration \n");
    printf("\n");
    printf("UPLINK\n");
    printf("  --hex [-x] Use hex input when using send_ack or send_unack\n");
    printf("  --send_ack {tx message string} Send acked uplink pkt\n");
    printf("  --send_unack {tx message string} Send un-acked uplink pkt\n");
    printf("\n");
    printf("DOWNLINK\n");
    printf("  --get_packet Get a packet after it has been received in downlink mode\n");
    printf("  --mailbox_request Sends a request to gateway for mailbox content\n");
    printf("\n");
    printf("STATUS\n");
    printf("  --state_get Get the current state of the module.\n");
    printf("  --net_info_get Get the network info\n");
    printf("  --irq_flags_get get state of IRQ Flags Register\n");
    printf("  --irq_flags_clear clear bits in IRQ Flags Register\n");
    printf("  --stats_get Get the communication statistics\n");
    printf("  --time_get Get the system time\n");
    printf("\n");
    printf("MODE AND RESET\n");
    printf("  --delete_settings [-d] delete saved settings from flash\n");
    printf("  --reset_mcu [-u] Reset the module (takes a few seconds)\n");
    printf("  --bootloader [-U] Reset to Bootloader mode (takes a few seconds)\n");
    printf("  --restore_defaults [-R] Restore default radio settings\n");
    printf("  --sleep [-s] Place module in sleep mode\n");
    printf("  --save_settings [-a] store radio settings to flash\n");
    printf("  --mac_set MAC mode set\n");
    printf("  --mac_get MAC mode get\n");
    printf("  --uuid get unique identifier\n");
#ifndef NEW_HOST_IFC
    printf("  --net_token_get Get the network token\n");
    printf("  --app_token_get Get the application token and registration status\n");
    printf("  --receive_get Get the current receive mode\n");
    printf("  --qos_get Get QoS level [0-15]\n");
#endif
    printf("  --time_sync Synchronize local time with the system time\n");
    printf("  --crypto_xchg_req Request a new encryption key exchange with gateway\n");
    printf("\n");
    printf("ADVANCED CONFIGURATION\n");
    printf("  --band_cfg_set Set the downlink frequency band\n");
    printf("  --band_cfg_get Get the downlink frequency band\n");

    printf("\n");
}

int main (int argc, char * const argv[])
{
    int c;
    int32_t i32_ret;    // Some function return int8_t's and some return int32_t's - know the difference!
    int8_t i8_ret;      // Some function return int8_t's and some return int32_t's - know the difference!

    /* Default options */
    int baudrate = LL_TTY_DEFAULT_BAUDRATE;
    char *dev_name = NULL;
    module_mode_t mode = MODE_NONE;
    char *buf = NULL;
    uint8_t delete_settings = 0;
    uint8_t store_settings = 0;
    uint8_t restore_defaults = 0;
    uint8_t reset_mcu = 0;
    uint8_t bootloader_mode = 0;
    uint8_t mac_set = SENTINEL;
    uint8_t mac_get = 0;
    uint8_t uuid = 0;

    uint8_t cmd_print_version = 0;
    uint8_t cmd_irq_flags_get = 0;
    uint8_t cmd_irq_flags_clear = 0;

    uint8_t cmd_net_token_set = 0;
    uint8_t cmd_app_token_set = 0;
    uint8_t cmd_receive_set = 0;
    uint8_t cmd_qos_request = SENTINEL;
    uint8_t cmd_config_get = 0;
#ifndef NEW_HOST_IFC
    uint8_t cmd_net_token_get = 0;
    uint8_t cmd_app_token_get = 0;
    uint8_t cmd_receive_get = 0;
    uint8_t cmd_qos_get = 0;
#endif
    uint32_t net_token = 0x4f50454e;
    //uint8_t app_token[10] = {0};
    uint32_t dl_mode = 0;
    uint32_t qos = 0;

    
    uint8_t cmd_crypto_xchg_req = 0;

    uint8_t cmd_mailbox_request = 0;
    uint8_t cmd_get_packet = 0;
    uint8_t use_hex = 0;

    uint8_t cmd_antenna_get = 0;
    uint8_t cmd_antenna_set = 0;
    uint8_t antenna_cfg;
    uint8_t cmd_state_get = 0;
    uint8_t cmd_net_info_get = 0;
    uint8_t cmd_stats_get = 0;
    uint8_t cmd_time_get = 0;
    uint8_t cmd_time_sync = 0;

    uint8_t cmd_band_cfg_get = 0;
    uint8_t cmd_band_cfg_set = 0;
    #define LEN_STR_BAND_CFG    (8*5+4+1)   // 5 values @ 8 chars each, 4 commas and a terminator
    char    str_band_cfg[LEN_STR_BAND_CFG] = {0};
    llabs_dl_band_cfg_t dl_band_cfg = {0};

    int len;


    /* Use getopt to parse command line arguments */
    while (1)
    {
        static struct option long_options[] =
        {
                {"version",             no_argument,       0, 'v'},
                {"baudrate",            required_argument, 0, 'b'},
                {"delete_settings",     no_argument,       0, 'd'},
                {"device",              required_argument, 0, 'D'},
                {"hex",                 no_argument,       0, 'x'},
                {"irq_flags_clear",     required_argument, 0,   0},
                {"irq_flags_get",       no_argument,       0,   0},
                {"crypto_xchg_req",     no_argument,       0,   0},
                {"net_token_set",       required_argument, 0,   0},
                {"app_token_set",       required_argument, 0,   0},
                {"receive_set",         required_argument, 0,   0},
                {"qos_set",             required_argument, 0,   0},
#ifndef NEW_HOST_IFC
                {"net_token_get",       no_argument,       0,   0},
                {"app_token_get",       no_argument,       0,   0},
                {"receive_get",         no_argument,       0,   0},
                {"qos_get",             no_argument,       0,   0},
#endif
                {"config_get",          no_argument,       0,   0},
                {"get_packet",          no_argument,       0,   0},
                {"mailbox_request",     no_argument,       0,   0},
                {"mac_set",             required_argument, 0,   0},
                {"mac_get",             no_argument,       0,   0},
                {"send_ack",            required_argument, 0,   0},
                {"send_unack",          required_argument, 0,   0},
                {"antenna_set",         required_argument, 0,   0},
                {"antenna_get",         no_argument,       0,   0},
                {"state_get",           no_argument,       0,   0},
                {"net_info_get",        no_argument,       0,   0},
                {"stats_get",           no_argument,       0,   0},
                {"time_get",            no_argument,       0,   0},
                {"time_sync",           no_argument,       0,   0},
                {"band_cfg_set",        required_argument, 0,   0},
                {"band_cfg_get",        no_argument,       0,   0},
                {"uuid",                no_argument,       0,   0},
                {"reset_mcu",           no_argument,       0, 'u'},
                {"bootloader",          no_argument,       0, 'U'},
                {"restore_defaults",    no_argument,       0, 'R'},
                {"save_settings",       no_argument,       0, 'a'},
                {"sleep",               no_argument,       0, 's'},
                {"help",                no_argument,       0, 'h'},
                {0, 0, 0, 0}
        };
        /* getopt_long stores the option index here. */
        int option_index = 0;

        c = getopt_long(argc, argv, "ab:dD:hkRsuUvx",
                        long_options, &option_index);

        /* Detect the end of the options. */
        if (c == -1)
        {
            break;
        }

        switch (c)
        {
            case 0:
                /* If this option set a flag, do nothing else now. */
                if (long_options[option_index].flag != 0)
                    break;
                if(strcmp(long_options[option_index].name, "mac_get") == 0)
                {
                    mac_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "mac_set") == 0)
                {
                    mac_set = (uint8_t)strtol(optarg, NULL, 0);
                }
                else if(strcmp(long_options[option_index].name, "uuid") == 0)
                {
                    uuid = 1;
                }
                else if(strcmp(long_options[option_index].name, "irq_flags_get") == 0)
                {
                    cmd_irq_flags_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "irq_flags_clear") == 0)
                {
                    buf = optarg;
                    cmd_irq_flags_clear = 1;
                }
                else if(strcmp(long_options[option_index].name, "crypto_xchg_req") == 0)
                {
                    cmd_crypto_xchg_req = 1;
                }
                else if(strcmp(long_options[option_index].name, "net_token_set") == 0)
                {
                    net_token = strtoul(optarg, NULL, 16);
                    cmd_net_token_set = 1;
                }
                else if(strcmp(long_options[option_index].name, "app_token_set") == 0)
                {
                    uint32_t len = strlen(optarg);
                    if(20 == len) 
                    {
                        buf = optarg;
                        cmd_app_token_set = 1;
                    }
                    else
                    {
                        printf("Incorrect Application Token format.\n");
                        cmd_app_token_set = 0;
                    }
                }
                else if(strcmp(long_options[option_index].name, "receive_set") == 0)
                {
                    dl_mode = (uint8_t)strtol(optarg, NULL, 0);
                    cmd_receive_set = 1;
                }
                else if(strcmp(long_options[option_index].name, "qos_set") == 0)
                {
                    qos = (uint8_t)strtol(optarg, NULL, 0);
                    cmd_qos_request = (uint8_t)strtol(optarg, NULL, 0);
                }
#ifndef NEW_HOST_IFC
                else if(strcmp(long_options[option_index].name, "net_token_get") == 0)
                {
                    cmd_net_token_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "app_token_get") == 0)
                {
                    cmd_app_token_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "receive_get") == 0)
                {
                    cmd_receive_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "qos_get") == 0)
                {
                    cmd_qos_get = 1;
                }
#endif
                else if(strcmp(long_options[option_index].name, "config_get") == 0)
                {
                    cmd_config_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "get_packet") == 0)
                {
                    cmd_get_packet = 1;
                }
                else if(strcmp(long_options[option_index].name, "mailbox_request") == 0)
                {
                    cmd_mailbox_request = 1;
                }
                else if(strcmp(long_options[option_index].name, "antenna_set") == 0)
                {
                    antenna_cfg = (uint8_t)strtol(optarg, NULL, 0);
                    cmd_antenna_set = 1;
                }
                else if(strcmp(long_options[option_index].name, "antenna_get") == 0)
                {
                    cmd_antenna_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "band_cfg_set") == 0)
                {
                    printf("optarg is :%s\n", optarg);
                    size_t optarg_len = strlen(optarg);
                    if (optarg_len > LEN_STR_BAND_CFG)
                    {
                        printf("Invalid length for band_cfg \n");
                    }
                    else
                    {
                        memcpy(str_band_cfg, optarg, optarg_len);
                        if (parse_band_cfg(str_band_cfg, &dl_band_cfg))
                        {
                            printf("Invalid band_cfg \n");
                        }
                        else
                        {
                            cmd_band_cfg_set = 1;
                        }
                    }
                }
                else if(strcmp(long_options[option_index].name, "band_cfg_get") == 0)
                {
                    cmd_band_cfg_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "state_get") == 0)
                {
                    cmd_state_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "net_info_get") == 0)
                {
                    cmd_net_info_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "stats_get") == 0)
                {
                    cmd_stats_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "time_get") == 0)
                {
                    cmd_time_get = 1;
                }
                else if(strcmp(long_options[option_index].name, "time_sync") == 0)
                {
                    cmd_time_sync = 1;
                }
                else if(strcmp(long_options[option_index].name, "send_ack") == 0)
                {
                    buf = optarg;
                    mode = MODE_SEND_ACK;
                }
                else if(strcmp(long_options[option_index].name, "send_unack") == 0)
                {
                    buf = optarg;
                    mode = MODE_SEND_UNACK;
                }
                else
                {
                    printf("option %s", long_options[option_index].name);
                }
                if (optarg)
                    printf(" with arg %s", optarg);
                printf("\n");
                break;

            case 'a':
                store_settings = 1;
                break;

            case 'b':
                baudrate = (int)strtol(optarg, NULL, 0);
                break;

            case 'd':
                delete_settings = 1;
                break;

            case 'D':
                dev_name = optarg;
                break;

            case 'h':
                g_print_help_exit = 1;
                break;

            case 'R':
                restore_defaults = 1;
                break;

            case 's':
                mode = MODE_SLEEP;
                break;

            case 'u':
                reset_mcu = 1;
                break;

            case 'U':
                bootloader_mode = 1;
                break;

            case 'v':
                cmd_print_version = 1;
                break;

            case 'x':
                use_hex = 1;
                break;

            case '?':
                /* getopt_long already printed an error message. */
                break;

            default:
                abort();
        }
    }
    if (g_print_help_exit)
    {
        usage();
        exit(EXIT_SUCCESS);
    }
    if(cmd_print_version)
    {
        printf("Executable Version: %d.%d.%d\n", VERSION_MAJOR, VERSION_MINOR, VERSION_TAG);
    }
    
    ll_tty_open(dev_name, baudrate);

    print_ll_firmware_type();
    print_ll_hardware_type();
    print_ll_version();
    print_ll_ifc_version();
    print_uuid();

    if(mac_get)
    {
        ll_mac_type_t mac_mode = MAC_INVALID;
        i32_ret = ll_mac_mode_get(&mac_mode);
        print_ll_ifc_error("ll_mac_mode_get", i32_ret);
        printf("MAC Mode = %d\n", mac_mode);
    }
    if(mac_set != SENTINEL)
    {
        i32_ret = ll_mac_mode_set(mac_set);
        print_ll_ifc_error("ll_mac_mode_set", i32_ret);
        printf("set MAC mode to %d\n", mac_set);
    }

    // Check that the module is in the correct mode for Symphony test
    {
        ll_mac_type_t mac_mode = MAC_INVALID;
        i32_ret = ll_mac_mode_get(&mac_mode);
        print_ll_ifc_error("check ll_mac_mode_get", i32_ret);
        if (SYMPHONY_LINK != mac_mode)
        {
            printf("WARNING: the module is in MAC mode (%d) instead of SYMPHONY_LINK mode (%d)\n", mac_mode, SYMPHONY_LINK);
        }
    }

    if(cmd_crypto_xchg_req)
    {
        printf("Requesting new encryption key exchange.\n");
        i32_ret = ll_encryption_key_exchange_request();
        print_ll_ifc_error("ll_encryption_key_exchange_request", i32_ret);
    }

#ifndef NEW_HOST_IFC
    if (cmd_net_token_set)
    {
        i32_ret = ll_net_token_set(net_token);
        print_ll_ifc_error("ll_net_token_set", i32_ret);
    }

    if (cmd_app_token_set)
    {
        printf("Application Token Request: %s\n", buf);

        // Convert from ASCII to integers
        int32_t len = buffer_to_hex(buf);

        // Feed to host_ifc
        i32_ret = ll_app_token_set((uint8_t*) buf, len);
        print_ll_ifc_error("ll_app_token_set", i32_ret);
    }

    if (cmd_receive_set)
    {
        i32_ret = ll_receive_mode_set(dl_mode);
        print_ll_ifc_error("ll_receive_mode_set", i32_ret);
    }

    if (cmd_qos_request != SENTINEL)
    {
        i32_ret = ll_qos_request(cmd_qos_request);
        print_ll_ifc_error("ll_qos_request", i32_ret);
        printf("QoS Request: %d\n", cmd_qos_request);
    }

    if (cmd_net_token_get)
    {
        uint32_t net_token;
        i32_ret = ll_net_token_get(&net_token);
        print_ll_ifc_error("ll_net_token_get", i32_ret);
        if (i32_ret >= 0)
        {
            printf("Network Token : %08x\n", net_token);
        }
    }
    if (cmd_app_token_get)
    {
        //get app_token
        uint8_t p[10] = {0};
        i32_ret = ll_app_token_get(p);
        print_ll_ifc_error("ll_app_token_get", i32_ret);
        if(i32_ret >= 0)
        {
            int32_t i;
            printf("Application Token is: ");
            for(i=0; i<i32_ret; i++)
            {
                printf("%02x", p[i]);
            }
            printf("\n");
        }

        //get app token registration status
        uint8_t is_registered = 0;
        i32_ret = ll_app_reg_get(&is_registered);
        print_ll_ifc_error("ll_app_reg_get", i32_ret);
        if(i32_ret >= 0)
        {
            if(0 == is_registered)
            {
                printf("NOT ");
            }
            printf("REGISTERED\n");
        }
    }

    if (cmd_receive_get)
    {
        uint8_t mode;
        int8_t ret = ll_receive_mode_get(&mode);
        print_ll_ifc_error("ll_receive_mode_get", ret);
        if (1 != ret)
        {
            printf("error getting rx mode: %d\n", ret);
        }
        else
        {
            printf("Rx mode: %d\n", mode);
        }
    }

    if (cmd_qos_get)
    {
        uint8_t qos = -1;
        i32_ret = ll_qos_get(&qos);
        print_ll_ifc_error("ll_qos_get", i32_ret);
        printf("QoS is %d\n", qos);
    }
#else
    // Symphony Configuration 
    if((0 != cmd_net_token_set) || (0 != cmd_app_token_set) || (0 != cmd_receive_set) || (SENTINEL != cmd_qos_request))
    {
        if((0 != cmd_net_token_set) && (0 != cmd_app_token_set) && (0 != cmd_receive_set) && (SENTINEL != cmd_qos_request))
        {
            // Convert from ASCII to integers
            int32_t len = buffer_to_hex(buf);
            if(10 == len)
            {
                uint8_t app_token[APP_TOKEN_LEN];
                memcpy(app_token, buf, 10);
                
                // OK to set configuration because all symphony params have been set
                i32_ret = ll_config_set(net_token, app_token, dl_mode, qos);
                print_ll_ifc_error("ll_config_set", i32_ret);
            }
            else
            {
                printf("ERROR: Invalid Application Token length: must be exactly 10 bytes.\n");
            }
        }
        else
        {
            printf("ERROR: Network Token, Application Token, Receive Mode, and QoS must all be set at the same time.\n");
            //return -1;
        }
    }
    
    if(cmd_config_get)
    {
        uint8_t app_token[APP_TOKEN_LEN] = {0};
        i32_ret = ll_config_get(&net_token, app_token, (enum ll_downlink_mode*) &dl_mode, (uint8_t*) &qos);
        print_ll_ifc_error("ll_config_get", i32_ret);
        
        if(0 <= i32_ret)
        {
            // Print Network Token
            printf("Network Token : %08x\n", net_token);

            // Print App Token
            int32_t i;
            printf("Application Token : ");
            for(i=0; i<10; i++)
            {
                printf("%02x", app_token[i]);
            }

            // Print App Token Registration
            uint8_t is_registered = 0;
            i32_ret = ll_app_reg_get(&is_registered);
            print_ll_ifc_error("ll_app_reg_get", i32_ret);
            if(i32_ret >= 0)
            {
                if(0 == is_registered)
                {
                    printf(" NOT");
                }
                printf(" REGISTERED\n");
            }
            
            // Print DL mode
            printf("Receive Mode : %d\n", dl_mode);
            
            // Print QoS
            printf("QoS : %d\n", qos);
        }
    }
#endif

    if (cmd_mailbox_request)
    {
        int32_t ret = ll_mailbox_request();
        print_ll_ifc_error("ll_mailbox_request", ret);
        if (0 < ret)
        {
            printf("Mailbox request sent \n");
        }
    }

    if (cmd_antenna_get)
    {
        uint8_t ant;
        i32_ret = ll_antenna_get(&ant);
        print_ll_ifc_error("ll_antenna_get", i32_ret);
        if (i32_ret >= 0)
        {
            printf("Antenna Configuration: %d\n", ant);
        }
    }

    if (cmd_antenna_set)
    {
        i32_ret = ll_antenna_set(antenna_cfg);
        print_ll_ifc_error("ll_receive_mode_set", i32_ret);
    }

    if (cmd_state_get)
    {
        enum ll_state state;
        enum ll_tx_state tx_state;
        enum ll_rx_state rx_state;

        i32_ret = ll_get_state(&state, &tx_state, &rx_state);
        print_ll_ifc_error("ll_get_state", i32_ret);
        if (i32_ret > 0)
        {
            printf("General state: ");
            switch (state)
            {
                case LL_STATE_IDLE_CONNECTED:
                    printf("Connected");
                    break;
                case LL_STATE_IDLE_DISCONNECTED:
                    printf("Disconnected");
                    break;
                case LL_STATE_INITIALIZING:
                    printf("Initializing");
                    break;
                case LL_STATE_ERROR:
                    printf("Error");
                    break;
            }
            printf("\n");

            printf("Tx state: ");
            switch (tx_state)
            {
                case LL_TX_STATE_TRANSMITTING:
                    printf("Transmitting");
                    break;
                case LL_TX_STATE_SUCCESS:
                    printf("Successful");
                    break;
                case LL_TX_STATE_ERROR:
                    printf("Unsuccessful");
                    break;
            }
            printf("\n");

            printf("Rx state: ");
            switch (rx_state)
            {
                case LL_RX_STATE_NO_MSG:
                    printf("No Message");
                    break;
                case LL_RX_STATE_RECEIVED_MSG:
                    printf("Message Received");
                    break;
            }
            printf("\n");
        }
    }

    if (cmd_net_info_get)
    {
        llabs_network_info_t net_info;

        i32_ret = ll_net_info_get(&net_info);
        print_ll_ifc_error("ll_net_info_get", i32_ret);
        if (i32_ret >= 0)
        {
            if(net_info.is_scanning_gateways)
            {
                printf("Scanning for Gateways... \n");
            }
            
            if (net_info.connection_status == 0)
            {
                printf("\nNo Network Info available.\n");
            }
            else
            {

                printf("\nNetwork Info:\n");
                printf("  Network Token (Node) : %X\n", net_info.network_id_node);
                printf("  Network Token (Gateway) : %X\n", net_info.network_id_gw);
                printf("  Gateway ID : $101$0-0-0-%09" PRIx64 "\n", net_info.gateway_id & 0xFFFFFFFFF);
                printf("  Gateway Channel : %d\n", net_info.gateway_channel);
                printf("  Gateway Frequency : %d [Hz]\n", net_info.gateway_frequency);
                printf("  Time Since Last Rx : %d [Seconds]\n", net_info.last_rx_tick);
                printf("  RSSI : %d [dBm]\n", net_info.rssi);
                printf("  SNR : %d [dB]\n", net_info.snr);
                printf("Connection Status: ");
                switch (net_info.connection_status)
                {
                    case 1:
                        printf("DISCONNECTED\n");
                        break;
                    case 2:
                        printf("CONNECTED\n");
                        break;
                    default:
                        printf("UNKNOWN (%d)\n", net_info.connection_status);
                        break;
                }
            }
        }
    }

    if (cmd_stats_get)
    {
        llabs_stats_t stats;
        i32_ret = ll_stats_get(&stats);
        print_ll_ifc_error("ll_stats_get", i32_ret);
        if (i32_ret >= 0)
        {
            printf("\nStats:\n");
            printf("  num_send_calls         = %d\n", stats.num_send_calls);
            printf("  num_pkts_transmitted   = %d\n", stats.num_pkts_transmitted);
            printf("  num_gateway_scans      = %d\n", stats.num_gateway_scans);
            printf("  num_collisions         = %d\n", stats.num_collisions);
            printf("  num_ack_successes      = %d\n", stats.num_ack_successes);
            printf("  num_ack_failures       = %d\n", stats.num_ack_failures);
            printf("  num_sync_failures      = %d\n", stats.num_sync_failures);
            printf("  num_canceled_pkts_ack  = %d\n", stats.num_canceled_pkts_ack);
            printf("  num_canceled_pkts_csma = %d\n", stats.num_canceled_pkts_csma);
            printf("  num_rx_errors          = %d\n\n", stats.num_rx_errors);
        }
    }

    if (cmd_time_get)
    {
        int cnt;
        for (cnt = 0; cnt < 40; cnt++)
        {
        llabs_time_info_t time_info;
        i32_ret = ll_system_time_get(&time_info);
            struct timeb timeb;
            ftime(&timeb);
        print_ll_ifc_error("ll_system_time_get", i32_ret);
        if (i32_ret >= 0)
        {
            char time_str[40];
            time_t time_sec;

            time_sec = time_info.curr.seconds;
            strftime(time_str, sizeof(time_str), "%Y-%m-%d %H:%M:%S", gmtime(&time_sec));
            printf("\nCurrent time: %s.%03d UTC\n", time_str, time_info.curr.millis);

            time_sec = time_info.last_sync.seconds;
            strftime(time_str, sizeof(time_str), "%Y-%m-%d %H:%M:%S", gmtime(&time_sec));
            printf("Last time sync: %s.%03d UTC\n", time_str, time_info.last_sync.millis);

            strftime(time_str, sizeof(time_str), "%Y-%m-%d %H:%M:%S", gmtime(&timeb.time));
            printf("Local time  : %s.%03d UTC\n", time_str, timeb.millitm);
            MS_SLEEP(50);
        }
        }
    }

    if (cmd_time_sync)
    {
        i32_ret = ll_system_time_sync();
        print_ll_ifc_error("ll_system_time_sync", i32_ret);
    }

    if (cmd_band_cfg_set)
    {
        i32_ret = ll_dl_band_cfg_set(&dl_band_cfg);
        print_ll_ifc_error("ll_dl_band_cfg_set", i32_ret);
    }

    if (cmd_band_cfg_get)
    {
        printf("\nDownlink Band Configuration:\n");
        i32_ret = ll_dl_band_cfg_get(&dl_band_cfg);
        print_ll_ifc_error("ll_dl_band_cfg_get", i32_ret);
        if (i32_ret >= 0)
        {
            print_band_cfg(&dl_band_cfg);
        }
    }

    if (cmd_get_packet)
    {
        uint8_t buff[255];
        uint8_t bytes_received;
        uint8_t i;
        int16_t rssi;
        uint8_t snr;

        if(ll_retrieve_message(buff, &bytes_received, &rssi, &snr))
        {
            printf("Error getting received packet\n");
        }
        else if (bytes_received > 0)
        {
            // print the received downlink message
            printf("Downlink Message|RSSI=%d|SNR=%d : 0x", rssi, snr);
            for (i = 0; i < bytes_received; i++)
            {
                printf("%02X", buff[i]);
            }
            printf("\n");
        }
        else
        {
            printf("No packets\n");
        }
    }

    if((cmd_irq_flags_get) || (cmd_irq_flags_clear))
    {
        uint32_t irq_flags_to_clear = 0;
        uint32_t irq_flags_read = 0;
        uint8_t len = 4;

        if (cmd_irq_flags_clear)
        {
            len = strlen(buf)/2;
            buffer_to_hex(buf);
            irq_flags_to_clear = ((uint32_t)buf[3]);
            irq_flags_to_clear |= ((uint32_t)buf[2]) << 8;
            irq_flags_to_clear |= ((uint32_t)buf[1]) << 16;
            irq_flags_to_clear |= ((uint32_t)buf[0]) << 24;
        }

        // printf("\tirq_flags_to_clear = 0x%08X\n", irq_flags_to_clear);

        if (len != 4)
        {
            printf("Argument must be 32-bit hex\n");
        }
        else
        {
            // Check what flags are set, and clear some if we want
            i8_ret = ll_irq_flags(irq_flags_to_clear, &irq_flags_read);
            if (i8_ret<0)
            {
                print_ll_ifc_error("ll_irq_flags", i8_ret);
            }
            else
            {
                if(cmd_irq_flags_get)
                {
                    printf("irq_flags = 0x%08X\n", irq_flags_read);
                    print_irq_flags_text(irq_flags_read);
                }
            }
        }
    }

    if (uuid)
    {
        print_uuid();
    }

    if (delete_settings)
    {
        i32_ret = ll_settings_delete();
        print_ll_ifc_error("ll_settings_delete", i32_ret);
    }
    if (store_settings)
    {
        i32_ret = ll_settings_store();
        print_ll_ifc_error("ll_settings_store", i32_ret);
    }
    if (restore_defaults)
    {
        i32_ret = ll_restore_defaults();
        print_ll_ifc_error("ll_restore_defaults", i32_ret);
    }

    if (reset_mcu)
    {
        printf("Resetting module\n");
        i32_ret = ll_reset_mcu();
        print_ll_ifc_error("ll_reset_mcu", i32_ret);
    }

    if (bootloader_mode)
    {
        printf("Putting module in bootloader mode\n");
        i32_ret = ll_bootloader_mode();
        print_ll_ifc_error("ll_bootloader_mode", i32_ret);
    }


    switch (mode)
    {
        case MODE_NONE:
            break;
        case MODE_SLEEP:
            i32_ret = ll_sleep();
            print_ll_ifc_error("ll_sleep", i32_ret);
            break;
        case MODE_SEND_ACK:
            if (use_hex)
            {
                len = strlen(buf)/2;
                buffer_to_hex(buf);
            }
            else
            {
                len = strlen(buf);
            }
            i32_ret = ll_packet_send_ack((uint8_t *)buf, len);
            print_ll_ifc_error("ll_packet_send_ack", i32_ret);
            printf("ll_packet_send_ack returned %d\n", i32_ret);
            break;
        case MODE_SEND_UNACK:
            if (use_hex)
            {
                len = strlen(buf)/2;
                buffer_to_hex(buf);
            }
            else
            {
                len = strlen(buf);
            }
            i32_ret = ll_packet_send_unack((uint8_t *)buf, len);
            print_ll_ifc_error("ll_packet_send_unack", i32_ret);
            printf("ll_packet_send_unack returned %d\n", i32_ret);
            break;
        default:
            fprintf(stderr, "ERROR: Invalid mode %d\n", mode);
    }

    //SLEEP(1);
    MS_SLEEP(100);

    return 0;
}
